# -*- coding: utf-8 -*-
"""
Provide data handling class and methods (storing, processing, plotting...)
"""
from ._core import *
from ._DataCollection_class import *
from ._spectrallines import *
from ._mesh import *




# Common
import numpy as np
import scipy.optimize as scpopt
import scipy.sparse as scpsparse
from scipy.interpolate import BSpline

# Temporary for debugging
import matplotlib.pyplot as plt


_JAC = 'dense'


###########################################################
###########################################################
#
#           Main function for fit1d
#
###########################################################
###########################################################


def multigausfit1d_from_dlines_funccostjac(
    lamb,
    indx=None,
    dinput=None,
    dind=None,
    scales=None,
    jac=None,
):

    if jac is None:
        jac = _JAC

    ibckax = dind['bck_amp']['x']
    ibckrx = dind['bck_rate']['x']
    nbck = 1    # ibckax.size + ibckrx.size
    iax = dind['amp']['x']
    iwx = dind['width']['x']
    ishx = dind['shift']['x']
    idratiox, idshx = None, None
    if dinput['double'] is not False:
        c0 = dinput['double'] is True
        if c0 or dinput['double'].get('dratio') is None:
            idratiox = dind['dratio']['x']
        if c0 or dinput['double'].get('dshift') is None:
            idshx = dind['dshift']['x']

    ial = dind['amp']['lines']
    iwl = dind['width']['lines']
    ishl = dind['shift']['lines']

    iaj = dind['amp']['jac']
    iwj = dind['width']['jac']
    ishj = dind['shift']['jac']

    coefsal = dinput['amp']['coefs']
    coefswl = dinput['width']['coefs']
    coefssl = dinput['shift']['coefs']

    offsetal = dinput['amp']['offset']
    offsetwl = dinput['width']['offset']
    offsetsl = dinput['shift']['offset']

    lambrel = lamb - np.nanmin(lamb)
    lambnorm = lamb[..., None]/dinput['lines'][None, ...]

    xscale = np.full((dind['sizex'],), np.nan)
    if indx is None:
        indx = np.ones((dind['sizex'],), dtype=bool)

    # # bsplines-specific
    # lambnormcost = lamb[indok].ravel()[:, None] / dinput['lines'][None, :]
    # libs = np.array([((phicost>=km[ii]) & (phicost<=km[ii+kpb-1]))
        # for ii in range(nbs)])
    # lbs = [BSpline.basis_element(
        # km[ii:ii+kpb],
        # extrapolate=False)(phicost[libs[ii]])[:, None]
        # for ii in range(nbs)]
    # BS = BSpline(km, np.ones(ial.shape, dtype=float), dinput['deg'],
        # extrapolate=False, axis=0)

    # lcond = [np.any(np.isnan(bs)) for bs in lbs]
    # if np.any(lcond):
        # msg = ("Some nan have been detected in the jacobian!\n"
        # + "\t- lbs[{}]".format(lcond.index(True)))
        # raise Exception(msg)

    # func_details returns result in same shape as input
    def func_detail(
        x,
        xscale=xscale,
        indx=indx,
        lambrel=lambrel,
        lambnorm=lambnorm,
        ibckax=ibckax,
        ibckrx=ibckrx,
        ial=ial,
        iwl=iwl,
        ishl=ishl,
        idratiox=idratiox,
        idshx=idshx,
        nlines=dinput['nlines'],
        nbck=nbck,
        coefsal=coefsal[None, :],
        coefswl=coefswl[None, :],
        coefssl=coefssl[None, :],
        offsetal=offsetal[None, :],
        offsetwl=offsetwl[None, :],
        offsetsl=offsetsl[None, :],
        double=dinput['double'],
        scales=None,
        indok=None,
        const=None,
    ):
        if indok is None:
            indok = np.ones(lamb.shape, dtype=bool)
        shape = tuple(np.r_[indok.sum(), nbck+nlines])
        y = np.full(shape, np.nan)
        xscale[indx] = x*scales[indx]
        xscale[~indx] = const

        # Prepare
        amp = xscale[ial]*coefsal + offsetal
        wi2 = xscale[iwl]*coefswl + offsetwl
        shift = xscale[ishl]*coefssl + offsetsl
        exp = np.exp(-(lambnorm[indok, :] - (1 + shift))**2 / (2*wi2))

        if double is not False:
            # coefssl are line-specific, they do not affect dshift
            if double is True:
                dratio = xscale[idratiox]
                dshift = shift + xscale[idshx]  # *coefssl
            else:
                dratio = double.get('dratio', xscale[idratiox])
                dshift = shift + double.get('dshift', xscale[idshx])
            expd = np.exp(-(lambnorm[indok, :] - (1 + dshift))**2 / (2*wi2))

        # compute y
        y[:, :nbck] = (
            xscale[ibckax]
            * np.exp(xscale[ibckrx]*lambrel[indok])
        )[:, None]
        y[:, nbck:] = amp * exp
        if double is not False:
            y[:, nbck:] += amp * dratio * expd
        return y

    # cost and jacob return flattened results (for least_squares())
    def cost(
        x,
        xscale=xscale,
        indx=indx,
        lambrel=lambrel,
        lambnorm=lambnorm,
        ibckax=ibckax,
        ibckrx=ibckrx,
        ial=ial,
        iwl=iwl,
        ishl=ishl,
        idratiox=idratiox,
        idshx=idshx,
        scales=scales,
        coefsal=coefsal[None, :],
        coefswl=coefswl[None, :],
        coefssl=coefssl[None, :],
        offsetal=offsetal[None, :],
        offsetwl=offsetwl[None, :],
        offsetsl=offsetsl[None, :],
        double=dinput['double'],
        indok=None,
        const=None,
        data=0.,
    ):
        if indok is None:
            indok = np.ones(lamb.shape, dtype=bool)

        # xscale = x*scales
        xscale[indx] = x*scales[indx]
        xscale[~indx] = const

        # make sure iwl is 2D to get all lines at once
        amp = xscale[ial] * coefsal + offsetal
        inv_2wi2 = 1./(2.*(xscale[iwl] * coefswl + offsetwl))
        shift = xscale[ishl] * coefssl + offsetsl
        y = (
            np.nansum(
                amp * np.exp(-(lambnorm[indok, :]-(1 + shift))**2 * inv_2wi2),
                axis=1,
            )
            + xscale[ibckax]*np.exp(xscale[ibckrx]*lambrel[indok])
        )
        if double is not False:
            if double is True:
                dratio = xscale[idratiox]
                # scales[ishl] or scales[idshx] ? coefssl ? +> no
                dshift = shift + xscale[idshx]
            else:
                dratio = double.get('dratio', xscale[idratiox])
                dshift = shift + double.get('dshift', xscale[idshx])

            y += np.nansum((amp * dratio
                            * np.exp(-(lambnorm[indok, :] - (1 + dshift))**2
                                     * inv_2wi2)), axis=1)

        if isinstance(data, np.ndarray):
            return y - data[indok]
        else:
            return y - data

    # Prepare jac
    if jac in ['dense', 'sparse']:
        # Define a callable jac returning (nlamb, sizex) matrix of partial
        # derivatives of np.sum(func_details(scaled), axis=0)

        # Pre-instanciate jacobian
        jac0 = np.zeros((lamb.size, dind['sizex']), dtype=float)
        # jac0 = np.zeros((lamb.size, indx.sum()), dtype=float)
        indxn = indx.nonzero()[0]

        def jacob(
            x,
            xscale=xscale,
            indx=indx,
            indxn=indxn,
            lambnorm=lambnorm,
            ibckax=ibckax,
            ibckrx=ibckrx,
            iax=iax, iaj=iaj, ial=ial,
            iwx=iwx, iwj=iwj, iwl=iwl,
            ishx=ishx, ishj=ishj, ishl=ishl,
            idratiox=idratiox, idshx=idshx,
            coefsal=coefsal[None, :],
            coefswl=coefswl[None, :],
            coefssl=coefssl[None, :],
            offsetal=offsetal[None, :],
            offsetwl=offsetwl[None, :],
            offsetsl=offsetsl[None, :],
            double=dinput['double'],
            scales=None, indok=None, data=None,
            jac0=jac0,
            const=None,
        ):
            """ Basic docstr """
            if indok is None:
                indok = np.ones(lamb.shape, dtype=bool)
            xscale[indx] = x*scales[indx]
            xscale[~indx] = const

            # Intermediates
            amp = xscale[ial] * coefsal + offsetal
            wi2 = xscale[iwl] * coefswl + offsetwl
            inv_wi2 = 1./wi2
            shift = xscale[ishl] * coefssl + offsetsl
            beta = (lambnorm[indok, :] - (1 + shift)) * inv_wi2 / 2.
            alpha = -beta**2 * (2*wi2)
            exp = np.exp(alpha)

            # Background
            jac0[indok, ibckax[0]] = (
                scales[ibckax[0]]
                * np.exp(xscale[ibckrx]*lambrel[indok])
            )
            jac0[indok, ibckrx[0]] = (
                xscale[ibckax[0]] * scales[ibckrx[0]] * lambrel[indok]
                * np.exp(xscale[ibckrx]*lambrel[indok])
            )

            # amp (shape: nphi/lamb, namp[jj])
            # for jj in range(len(iaj)):
            for jj, aa in enumerate(iaj):
                jac0[indok, iax[jj]] = np.sum(
                    exp[:, aa] * coefsal[:, aa],
                    axis=1) * scales[iax[jj]]

            # width2
            for jj in range(len(iwj)):
                jac0[indok, iwx[jj]] = np.sum(
                    (-alpha[:, iwj[jj]] * amp[:, iwj[jj]]
                     * exp[:, iwj[jj]] * coefswl[:, iwj[jj]]
                     * inv_wi2[:, iwj[jj]]), axis=1) * scales[iwx[jj]]

            # shift
            for jj in range(len(ishj)):
                jac0[indok, ishx[jj]] = np.sum(
                    (amp[:, ishj[jj]] * 2. * beta[:, ishj[jj]])
                    * exp[:, ishj[jj]] * coefssl[:, ishj[jj]],
                    axis=1) * scales[ishx[jj]]

            # double
            if double is not False:

                if double is True:
                    dratio = xscale[idratiox]
                    # coefssl are line-specific, they do not affect dshift
                    dshift = shift + xscale[idshx]
                else:
                    dratio = double.get('dratio', xscale[idratiox])
                    dshift = shift + double.get('dshift', xscale[idshx])
                # ampd not defined to save memory => *dratio instead
                # ampd = amp*dratio
                betad = (lambnorm[indok, :] - (1 + dshift)) * inv_wi2 / 2.
                alphad = -betad**2 * (2*wi2)
                expd = np.exp(alphad)

                # amp
                for jj in range(len(iaj)):
                    jac0[indok, iax[jj]] += dratio*np.sum(
                        expd[:, iaj[jj]] * coefsal[:, iaj[jj]],
                        axis=1) * scales[iax[jj]]

                # width2
                for jj in range(len(iwj)):
                    jac0[indok, iwx[jj]] += np.sum(
                        (
                            -alphad[:, iwj[jj]] * amp[:, iwj[jj]]
                            * expd[:, iwj[jj]] * coefswl[:, iwj[jj]]
                            * inv_wi2[:, iwj[jj]]
                        ),
                        axis=1,
                    ) * scales[iwx[jj]] * dratio

                # shift
                for jj in range(len(ishj)):
                    jac0[indok, ishx[jj]] += np.sum(
                        (amp[:, ishj[jj]] * 2.*betad[:, ishj[jj]]
                         * expd[:, ishj[jj]] * coefssl[:, ishj[jj]]),
                        axis=1) * scales[ishx[jj]] * dratio

                # dratio
                if double is True or double.get('dratio') is None:
                    jac0[indok, idratiox] = (
                        scales[idratiox] * np.sum(amp * expd, axis=1)
                    )

                # dshift
                if double is True or double.get('dshift') is None:
                    jac0[indok, idshx] = dratio * np.sum(
                        amp * 2.*betad*scales[idshx] * expd, axis=1)

            return jac0[indok, :][:, indx]

    elif jac == 'sparse':
        msg = "Sparse jacobian is pointless for 1d spectrum fitting"
        raise Exception(msg)

    if jac not in ['dense', 'sparse', 'LinearSparseOperator']:
        if jac not in ['2-point', '3-point']:
            msg = ("jac should be in "
                   + "['dense', 'sparse', 'LinearsparseOp', "
                   + "'2-point', '3-point']\n"
                   + "\t- provided: {}".format(jac))
            raise Exception(msg)
        jacob = jac

    return func_detail, cost, jacob

###########################################################
###########################################################
#
#           Main function for fit2d
#
###########################################################
###########################################################


def multigausfit2d_from_dlines_funccostjac(lamb, phi,
                                           dinput=None,
                                           binning=None,
                                           dind=None,
                                           scales=None,
                                           indok=None,
                                           jac=None):

    if jac is None:
        jac = _JAC

    ibckx = dind['bck']['x']
    iax = dind['amp']['x']
    iwx = dind['width']['x']
    ishx = dind['shift']['x']
    idratiox, idshx = None, None
    if dinput['double'] is not False:
        c0 = dinput['double'] is True
        if c0 or dinput['double'].get('dratio') is None:
            idratiox = dind['dratio']['x']
        if c0 or dinput['double'].get('dshift') is None:
            idshx = dind['dshift']['x']

    ial = dind['amp']['lines']
    iwl = dind['width']['lines']
    ishl = dind['shift']['lines']

    iaj = dind['amp']['jac']
    iwj = dind['width']['jac']
    ishj = dind['shift']['jac']

    coefsal = dinput['amp']['coefs']
    coefswl = dinput['width']['coefs']
    coefssl = dinput['shift']['coefs']

    offsetal = dinput['amp']['offset']
    offsetwl = dinput['width']['offset']
    offsetsl = dinput['shift']['offset']

    # If no binning => no fast indexing on lamb / phi (e.g.: for sparse)
    km = dinput['knots_mult']
    kpb = dinput['nknotsperbs']
    nbs = dinput['nbs']

    # bsplines-specific
    phicost = phi[indok].ravel()
    lambnormcost = lamb[indok].ravel()[:, None] / dinput['lines'][None, :]
    libs = np.array([((phicost >= km[ii]) & (phicost <= km[ii+kpb-1]))
                     for ii in range(nbs)])
    lbs = [BSpline.basis_element(
        km[ii:ii+kpb],
        extrapolate=False)(phicost[libs[ii]])[:, None]
           for ii in range(nbs)]
    BS = BSpline(km, np.ones(ial.shape, dtype=float), dinput['deg'],
                 extrapolate=False, axis=0)

    lcond = [np.any(np.isnan(bs)) for bs in lbs]
    if np.any(lcond):
        msg = ("Some nan have been detected in the jacobian!\n"
               + "\t- lbs[{}]".format(lcond.index(True)))
        raise Exception(msg)

    # func_details returns result in same shape as input
    def func_detail(x, phi=phi,
                    lambnorm=lamb[..., None]/dinput['lines'][None, ...],
                    ibckx=ibckx,
                    ial=ial,
                    iwl=iwl,
                    ishl=ishl,
                    idratiox=idratiox,
                    idshx=idshx,
                    nlines=dinput['nlines'],
                    km=km, kpb=kpb, nbs=nbs,
                    deg=dinput['deg'],
                    BS=BS,
                    nbck=int(ibckx.size/dinput['nbs']),
                    coefsal=coefsal[None, :],
                    coefswl=coefswl[None, :],
                    coefssl=coefssl[None, :],
                    double=dinput['double'],
                    scales=None,
                    indok=None,
                    indbs=None):
        if indok is None:
            indok = np.ones(phi.shape, dtype=bool)
        shape = tuple(np.r_[phi.shape, nbck+nlines, nbs])
        y = np.full(shape, np.nan)
        xscale = x*scales
        # make sure iwl is 2D to get all lines at once
        BS.c = xscale[iwl] * coefswl
        wi2 = BS(phi)
        BS.c = xscale[ishl] * coefssl
        shift = BS(phi)
        exp = np.exp(-(lambnorm - (1 + shift))**2 / (2*wi2))

        if double is not False:
            # coefssl are line-specific, they do not affect dshift
            if double is True:
                dratio = xscale[idratiox]
                dshift = shift + xscale[idshx]  # *coefssl
            else:
                dratio = double.get('dratio', xscale[idratiox])
                dshift = shift + double.get('dshift', xscale[idshx])
            expd = np.exp(-(lambnorm - (1 + dshift))**2 / (2*wi2))

        # Loop on individual bsplines for amp
        for ii in range(nbs):
            bs = BSpline.basis_element(km[ii:ii+kpb],
                                       extrapolate=False)(phi)

            indbs = ~np.isnan(bs)
            bs = bs[indbs]
            y[indbs, 0, ii] = xscale[ibckx[ii]]*bs
            for jj in range(nlines):
                amp = bs * xscale[ial[ii, jj]] * coefsal[0, jj]
                y[indbs, nbck+jj, ii] = amp * exp[indbs, jj]
                if double is not False:
                    y[indbs, nbck+jj, ii] += (amp * dratio * expd[indbs, jj])
        return y

    # cost and jacob return flattened results (for least_squares())
    def cost(x, phi=phicost,
             lambnorm=lambnormcost,
             ibckx=ibckx,
             ial=ial,
             iwl=iwl,
             ishl=ishl,
             idratiox=idratiox,
             idshx=idshx,
             km=km, kpb=kpb,
             deg=dinput['deg'],
             scales=scales,
             coefsal=coefsal[None, :],
             coefswl=coefswl[None, :],
             coefssl=coefssl[None, :],
             double=dinput['double'],
             indok_var=None, ind_bs=None, data=0.):

        xscale = x*scales

        # Background
        y = BSpline(km, xscale[ibckx], deg,
                    extrapolate=False, axis=0)(phi)

        # make sure iwl is 2D to get all lines at once
        amp = BSpline(km, xscale[ial] * coefsal, deg,
                      extrapolate=False, axis=0)(phi)
        wi2 = BSpline(km, xscale[iwl] * coefswl, deg,
                      extrapolate=False, axis=0)(phi)
        csh = xscale[ishl] * coefssl
        y += np.nansum((amp
                        * np.exp(-(lambnormcost
                                   - (1 + BSpline(km, csh, deg,
                                                  extrapolate=False,
                                                  axis=0)(phi)))**2
                                 / (2*wi2))),
                       axis=1)
        import pdb; pdb.set_trace()    # DB
        if double is not False:
            if double is True:
                dratio = xscale[idratiox]
                # scales[ishl] or scales[idshx] ? coefssl ? +> no
                csh = csh + xscale[idshx]
            else:
                dratio = double.get('dratio', xscale[idratiox])
                csh = csh + double.get('dshift', xscale[idshx])

            y += np.nansum((amp * dratio
                            * np.exp(-(lambnormcost
                                       - (1 + BSpline(
                                           km, csh, deg,
                                           extrapolate=False,
                                           axis=0)(phi)))**2
                                     / (2*wi2))),
                           axis=1)
        return y - data

    # Prepare jac
    if jac in ['dense', 'sparse']:
        # Define a callable jac returning (nlamb, sizex) matrix of partial
        # derivatives of np.sum(func_details(scaled), axis=0)

        # Pre-instanciate jacobian
        jac0 = np.zeros((phicost.size, dind['sizex']), dtype=float)

        def jacob(x, lambnorm=lambnormcost, phi=phicost,
                  ibckx=ibckx,
                  iax=iax, iaj=iaj, ial=ial,
                  iwx=iwx, iwj=iwj, iwl=iwl,
                  ishx=ishx, ishj=ishj, ishl=ishl,
                  idratiox=idratiox, idshx=idshx,
                  km=km, kpb=kpb, nbs=nbs,
                  coefsal=coefsal[None, :],
                  coefswl=coefswl[None, :],
                  coefssl=coefssl[None, :],
                  double=dinput['double'],
                  scales=None, indok_var=None, data=None,
                  jac0=jac0, lbs=lbs, libs=libs, BS=BS):
            """ Basic docstr """
            xscale = x*scales

            # Intermediates

            # Loop on bs
            for ii in range(nbs):
                bs = lbs[ii]

                # Intermediates
                BS.c = xscale[ial] * coefsal
                amp = BS(phi[libs[ii]])
                BS.c = xscale[iwl] * coefswl
                wi2 = BS(phi[libs[ii]])
                BS.c = xscale[ishl] * coefssl
                shift = BS(phi[libs[ii]])
                beta = (lambnorm[libs[ii], :] - (1 + shift)) / (2*wi2)
                alpha = -beta**2 * (2*wi2)
                # exp = np.exp(alpha)
                bsexp = bs * np.exp(alpha)

                # Background
                jac0[libs[ii], ibckx[ii]] = bs[:, 0] * scales[ibckx[ii]]

                # amp (shape: nphi/lamb, namp[jj])
                for jj in range(len(iaj)):
                    ix = iax[ii, jj]
                    # jac0[libs[ii], ix] = np.sum(
                    # (bs * exp[:, iaj[jj]] * scales[ix]
                    # * coefsal[0:1, iaj[jj]]),
                    # axis=1)
                    jac0[libs[ii], ix] = np.sum(
                        bsexp[:, iaj[jj]] * coefsal[:, iaj[jj]],
                        axis=1) * scales[ix]

                # width2
                for jj in range(len(iwj)):
                    ix = iwx[ii, jj]
                    # jac0[libs[ii], ix] = np.sum(
                    # (amp[:, iwj[jj]]
                    # * (-alpha[:, iwj[jj]]
                    # * bs * exp[:, iwj[jj]] * scales[ix]
                    # * coefswl[0:1, iwj[jj]]
                    # / wi2[:, iwj[jj]])),
                    # axis=1)
                    jac0[libs[ii], ix] = np.sum(
                        (
                            -alpha[:, iwj[jj]] * amp[:, iwj[jj]]
                            * bsexp[:, iwj[jj]] * coefswl[:, iwj[jj]]
                            / wi2[:, iwj[jj]]
                        ),
                        axis=1,
                    ) * scales[ix]

                # shift
                for jj in range(len(ishj)):
                    ix = ishx[ii, jj]
                    # jac0[libs[ii], ix] = np.sum(
                    # (amp[:, ishj[jj]]
                    # * 2.*beta[:, ishj[jj]]
                    # * bs * exp[:, ishj[jj]] * scales[ix]
                    # * coefssl[0:1, ishj[jj]]),
                    # axis=1)
                    jac0[libs[ii], ix] = np.sum(
                        (
                            amp[:, ishj[jj]] * 2. * beta[:, ishj[jj]]
                            * bsexp[:, ishj[jj]] * coefssl[:, ishj[jj]]
                        ),
                        axis=1,
                    ) * scales[ix]

                # double
                if double is False:
                    continue

                if double is True:
                    dratio = xscale[idratiox]
                    # coefssl are line-specific, they do not affect dshift
                    dshift = shift + xscale[idshx]
                else:
                    dratio = double.get('dratio', xscale[idratiox])
                    dshift = shift + double.get('dshift', xscale[idshx])

                # ampd = amp*dratio
                betad = (lambnorm[libs[ii], :] - (1 + dshift)) / (2*wi2)
                alphad = -betad**2 * (2*wi2)
                expd = np.exp(alphad)
                bsexpd = bs * expd

                # amp
                for jj in range(len(iaj)):
                    ix = iax[ii, jj]
                    # jac0[libs[ii], ix] += dratio*np.sum(
                    # (bs * scales[ix] * coefsal[0:1, iaj[jj]]
                    # * expd[:, iaj[jj]]),
                    # axis=1)
                    jac0[libs[ii], ix] += dratio*np.sum(
                        bsexpd[:, iaj[jj]] * coefsal[:, iaj[jj]],
                        axis=1) * scales[ix]

                # width2
                for jj in range(len(iwj)):
                    ix = iwx[ii, jj]
                    # jac0[libs[ii], ix] += np.sum(
                    # (ampd[:, iwj[jj]]
                    # * (-alphad[:, iwj[jj]]
                    # * bs * scales[ix] * coefswl[0:1, iwj[jj]]
                    # / wi2[:, iwj[jj]])
                    # * expd[:, iwj[jj]]),
                    # axis=1)
                    jac0[libs[ii], ix] += np.sum(
                        (-alphad[:, iwj[jj]] * amp[:, iwj[jj]]
                         * bsexpd[:, iwj[jj]] * coefswl[:, iwj[jj]]
                         / wi2[:, iwj[jj]]), axis=1) * scales[ix] * dratio

                # shift
                for jj in range(len(ishj)):
                    ix = ishx[ii, jj]
                    # jac0[libs[ii], ix] += np.sum(
                    # (ampd[:, ishj[jj]]
                    # * 2.*betad[:, ishj[jj]]
                    # * bs * scales[ix] * coefssl[0:1, ishj[jj]]
                    # * expd[:, ishj[jj]]),
                    # axis=1)
                    jac0[libs[ii], ix] += np.sum(
                        (amp[:, ishj[jj]] * 2.*betad[:, ishj[jj]]
                         * bsexpd[:, ishj[jj]] * coefssl[:, ishj[jj]]),
                        axis=1) * scales[ix] * dratio

                # dratio
                if double is True or double.get('dratio') is None:
                    jac0[libs[ii], idratiox] = (scales[idratiox]
                                                * np.sum(amp * expd, axis=1))

                # dshift
                if double is True or double.get('dshift') is None:
                    jac0[libs[ii], idshx] = dratio * np.sum(
                        amp * 2.*betad*scales[idshx] * expd, axis=1)
            return jac0

    elif jac == 'sparse':

        msg = ""
        raise Exception(msg)

        # Define a callable jac returning (nlamb, sizex) matrix of partial
        # derivatives of np.sum(func_details(scaled), axis=0)

        # Pre-instanciate jacobian
        # libsT = libs.T
        # indflat = np.flatnonzero(libsT)
        # row_ind, col_ind = libsT.nonzero()
        # libsi = [indflat[col_ind==ii] for ii in range(nbs)]

        # jac0 = scpsparse.csr_matrix((libsT[libsT].astype(float),
        # (row_ind, col_ind)),
        # shape=libsT.shape)

        # def jacob(x, lamb=lambcost, phi=phicost,
        # ibckx=ibckx,
        # iax=iax, iaj=iaj, ial=ial,
        # iwx=iwx, iwj=iwj, iwl=iwl,
        # ishx=ishx, ishj=ishj, ishl=ishl,
        # idratiox=idratiox, idshx=idshx,
        # lines=dinput['lines'][None, :],
        # km=km, kpb=kpb, nbs=nbs,
        # deg=dinput['deg'],
        # coefsal=coefsal[None, :],
        # coefswl=coefswl[None, :],
        # coefssl=coefssl[None, :],
        # double=dinput['double'],
        # scales=None, indok_var=None, data=None,
        # jac0=jac0, lbs=lbs, libsi=libsi):
        # """ Basic docstr """
        # xscale = x*scales

        # # Intermediates

    # # Loop on bs
    # for ii in range(nbs):
        # bs = lbs[ii]

        # # Intermediates
        # amp = BSpline(km, xscale[ial] * coefsal, deg,
        # extrapolate=False, axis=0)(phi[libs[ii]])
        # wi2 = BSpline(km, xscale[iwl] * coefswl, deg,
        # extrapolate=False, axis=0)(phi[libs[ii]])
        # shift = BSpline(km, xscale[ishl] * coefssl, deg,
        # extrapolate=False, axis=0)(phi[libs[ii]])
        # beta = (lamb[libs[ii], :]/lines - (1 + shift)) / (2*wi2)
        # alpha = -beta**2 * (2*wi2)
        # exp = np.exp(alpha)

        # # Background
        # import ipdb; ipdb.set_trace()       # DB
        # jac0.data[libsi[ibckx[ii]]] = bs[:, 0] * scales[ibckx[ii]]
        # #jac0[libs[ii], ibckx[ii]] = bs[:, 0] * scales[ibckx[ii]]

        # # amp
        # for jj in range(len(iaj)):
        # ix = iax[ii, jj]
        # jac0.data[libs[ii], ix] = np.sum(
        # (bs * exp[:, iaj[jj]] * scales[ix]
        # * coefsal[0:1, iaj[jj]]),
        # axis=1)

        # # width2
        # for jj in range(len(iwj)):
        # ix = iwx[ii, jj]
        # jac0[libs[ii], ix] = np.sum(
        # (amp[:, iwj[jj]]
        # * (-alpha[:, iwj[jj]]
        # * bs * exp[:, iwj[jj]] * scales[ix]
        # * coefswl[0:1, iwj[jj]]
        # / wi2[:, iwj[jj]])),
        # axis=1)

        # # shift
        # for jj in range(len(ishj)):
        # ix = ishx[ii, jj]
        # jac0[libs[ii], ix] = np.sum(
        # (amp[:, ishj[jj]]
        # * 2.*beta[:, ishj[jj]]
        # * bs * exp[:, ishj[jj]] * scales[ix]
        # * coefssl[0:1, ishj[jj]]),
        # axis=1)

        # # double
        # if double is False:
        # continue

        # if double is True:
        # dratio = xscale[idratiox]
        # # coefssl are line-specific, they do not affect dshift
        # shiftd = shift + xscale[idshx]
        # else:
        # if double.get('dratio') is None:
        # dratio = xscale[idratiox]
        # else:
        # dratio = double.get('dratio')
        # if double.get('dshift') is None:
        # shiftd = shift + xscale[idshx]
        # else:
        # shiftd = shift + double.get('dshift')

        # ampd = amp*dratio
        # betad = (lamb[libs[ii]]/lines - (1 + shiftd)) / (2*wi2)
        # alphad = -betad**2 * (2*wi2)
        # expd = np.exp(alphad)

        # # amp
        # for jj in range(len(iaj)):
        # ix = iax[ii, jj]
        # jac0[libs[ii], ix] += dratio*np.sum(
        # (bs * scales[ix] * coefsal[0:1, iaj[jj]]
        # * expd[:, iaj[jj]]),
        # axis=1)

        # # width2
        # for jj in range(len(iwj)):
        # ix = iwx[ii, jj]
        # jac0[libs[ii], ix] += np.sum(
        # (ampd[:, iwj[jj]]
        # * (-alphad[:, iwj[jj]]
        # * bs * scales[ix] * coefswl[0:1, iwj[jj]]
        # / wi2[:, iwj[jj]])
        # * expd[:, iwj[jj]]),
        # axis=1)

        # # shift
        # for jj in range(len(ishj)):
        # ix = ishx[ii, jj]
        # jac0[libs[ii], ix] += np.sum(
        # (ampd[:, ishj[jj]]
        # * 2.*betad[:, ishj[jj]]
        # * bs * scales[ix] * coefssl[0:1, ishj[jj]]
        # * expd[:, ishj[jj]]),
        # axis=1)

        # # dratio
        # if double is True or double.get('dratio') is None:
        # jac0[libs[ii], idratiox] = (scales[idratiox]
        # *np.sum(amp*expd, axis=1))

        # # dshift
        # if double is True or double.get('dshift') is None:
        # jac0[libs[ii], idshx] = np.sum(
        # ampd * 2.*betad*scales[idshx] * expd, axis=1)
        # return jac0[indok, :]

    if jac not in ['dense', 'sparse', 'LinearSparseOperator']:
        if jac not in ['2-point', '3-point']:
            msg = ("jac should be in "
                   + "['dense', 'sparse', 'LinearsparseOp', "
                   + "'2-point', '3-point']\n"
                   + "\t- provided: {}".format(jac))
            raise Exception(msg)
        jacob = jac

    return func_detail, cost, jacob



# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np


# tofu
# from tofu import __version__ as __version__
from . import _generic_check
from ._DataCollection_class import DataCollection
from . import _mesh_checks
from . import _mesh_comp
from . import _mesh_plot
from . import _matrix_comp
from . import _matrix_plot


_GROUP_MESH = 'mesh'
_GROUP_R = 'R'
_GROUP_Z = 'Z'


# #############################################################################
# #############################################################################
#                           Mesh2DRect
# #############################################################################


class Mesh2DRect(DataCollection):

    _ddef = {
        'Id': {'include': ['Mod', 'Cls', 'Name', 'version']},
        'params': {
            'lambda0': (float, 0.),
            'source': (str, 'unknown'),
            'transition':    (str, 'unknown'),
            'element':  (str, 'unknown'),
            'charge':  (int, 0),
            'ion':  (str, 'unknown'),
            'symbol':   (str, 'unknown'),
        },
    }
    _forced_group = [_GROUP_R, _GROUP_Z]
    _data_none = True

    _show_in_summary_core = ['shape', 'ref', 'group']
    _show_in_summary = 'all'

    _groupmesh = _GROUP_MESH
    _groupR = _GROUP_R
    _groupZ = _GROUP_Z

    def add_mesh(
        self,
        key=None,
        domain=None,
        res=None,
    ):
        """ Add a mesh by key

        """

        dref, dmesh = _mesh_checks._mesh2DRect_to_dict(
            domain=domain,
            res=res,
            key=key,
        )
        dobj = {
            self._groupmesh: dmesh,
        }
        self.update(dref=dref, dobj=dobj)

    # -----------------
    # from config
    # ------------------

    @classmethod
    def from_Config(
        cls,
        config=None,
        key_struct=None,
        res=None,
        deg=None,
        key=None,
        thresh_in=None,
    ):
        """

        Example:
        --------
                >>> import tofu as tf
                >>> conf = tf.load_config('ITER')
                >>> mesh = tf.data.Mesh2DRect.from_Config(
                    config=conf,
                    res=[],
                )

        """

        domain, poly = _mesh_checks._mesh2DRect_from_Config(
            config=config, key_struct=key_struct,
        )

        obj = cls()
        obj.add_mesh(domain=domain, res=res, key=key)
        if deg is not None:
            obj.add_bsplines(deg=deg)
        obj.crop(key=key, crop=poly, thresh_in=thresh_in)
        return obj

    # -----------------
    # bsplines
    # ------------------

    def add_bsplines(self, key=None, deg=None):
        """ Add bspline basis functions on the chosen mesh """

        # --------------
        # check inputs

        keym, keybs, deg = _mesh_checks._mesh2DRect_bsplines(
            key=key,
            lkeys=list(self.dobj[self._groupmesh].keys()),
            deg=deg,
        )

        # --------------
        # get bsplines

        dref, dobj = _mesh_comp._mesh2DRect_bsplines(
            mesh=self, keym=keym, keybs=keybs, deg=deg,
        )

        self.update(dobj=dobj, dref=dref)
        _mesh_comp.add_cropbs_from_crop(mesh=self, keybs=keybs, keym=keym)

    # -----------------
    # get data subset
    # ------------------

    def get_profiles2d(self):
        """ Return dict of profiles2d with associated bsplines as values """

        # dict of bsplines shapes
        dbs = {
            k0: v0['ref']
            for k0, v0 in self.dobj['bsplines'].items()
        }

        # dict of profiles2d
        dk = {
            k0: [k1 for k1, v1 in dbs.items() if v0['ref'][-2:] == v1][0]
            for k0, v0 in self.ddata.items()
            if len([k1 for k1, v1 in dbs.items() if v0['ref'][-2:] == v1]) == 1
        }
        dk.update({k0: k0 for k0 in dbs.keys()})

        return dk

    # -----------------
    # indices
    # ------------------

    def select_ind(
        self,
        key=None,
        ind=None,
        elements=None,
        returnas=None,
        crop=None,
    ):
        """ Return ind for selected key (mesh or bspline) as:
                - tuple (default)
                - 'flat'

        Can covert one into the other
        """
        return _mesh_comp._select_ind(
            mesh=self,
            key=key,
            ind=ind,
            elements=elements,
            returnas=returnas,
            crop=crop,
        )

    def select_mesh_elements(
        self,
        key=None,
        ind=None,
        elements=None,
        returnas=None,
        return_neighbours=None,
        crop=None,
    ):
        """ Return indices or values of selected knots / cent

        Can be used to convert tuple (R, Z) indices to flat (RZ,) indices
        Can return values instead of indices
        Can return indices / values of neighbourgs

        """
        lk = list(self.dobj[self._groupmesh].keys())
        if key is None and len(lk) == 1:
            key = lk[0]
        ind = self.select_ind(
            key=key, ind=ind, elements=elements, returnas=tuple, crop=crop,
        )
        return _mesh_comp._select_mesh(
            mesh=self,
            key=key,
            ind=ind,
            elements=elements,
            returnas=returnas,
            return_neighbours=return_neighbours,
        )

    def select_bsplines(
        self,
        key=None,
        ind=None,
        returnas=None,
        return_cents=None,
        return_knots=None,
        crop=None,
    ):
        """ Return indices or values of selected knots / cent

        Can be used to convert tuple (R, Z) indices to flat (RZ,) indices
        Can return values instead of indices
        Can return indices / values of neighbourgs

        """
        lk = list(self.dobj['bsplines'].keys())
        if key is None and len(lk) == 1:
            key = lk[0]
        ind = self.select_ind(key=key, ind=ind, returnas=tuple, crop=crop)
        return _mesh_comp._select_bsplines(
            mesh=self,
            key=key,
            ind=ind,
            returnas=returnas,
            return_cents=return_cents,
            return_knots=return_knots,
        )

    # -----------------
    # Integration operators
    # ------------------

    def add_bsplines_operator(
        self,
        key=None,
        operator=None,
        geometry=None,
        crop=None,
        store=None,
        returnas=None,
    ):
        """ Get a matrix operator to compute an integral

        operator specifies the integrand:
            - 'D0': integral of the value
            - 'D0N2': integral of the squared value
            - 'D1N2': integral of the squared gradient
            - 'D2N2': integral of the squared laplacian

        geometry specifies in which geometry:
            - 'linear': linear geometry (cross-section = surface)
            - 'toroidal': toroildal geometry (cross-section = volumic slice)

        """

        (
            opmat, operator, geometry, dim, ref, crop,
            store, returnas, key,
        ) = _mesh_comp.get_bsplines_operator(
            self,
            key=key,
            operator=operator,
            geometry=geometry,
            crop=crop,
            store=store,
            returnas=returnas,
        )

        # store
        if store is True:
            if operator in ['D0', 'D0N2']:
                name = f'{key}-{operator}-{geometry}'
                self.add_data(
                    key=name,
                    data=opmat,
                    ref=ref,
                    units='',
                    name=operator,
                    dim=dim,
                )
            elif operator == 'D1N2':
                name = f'{key}-{operator}-dR-{geometry}'
                self.add_data(
                    key=name,
                    data=opmat[0],
                    ref=ref,
                    units='',
                    name=operator,
                    dim=dim,
                )
                name = f'{key}-{operator}-dZ-{geometry}'
                self.add_data(
                    key=name,
                    data=opmat[1],
                    ref=ref,
                    units='',
                    name=operator,
                    dim=dim,
                )
            elif operator == 'D2N2':
                name = f'{key}-{operator}-d2R-{geometry}'
                self.add_data(
                    key=name,
                    data=opmat[0],
                    ref=ref,
                    units='',
                    name=operator,
                    dim=dim,
                )
                name = f'{key}-{operator}-d2Z-{geometry}'
                self.add_data(
                    key=name,
                    data=opmat[1],
                    ref=ref,
                    units='',
                    name=operator,
                    dim=dim,
                )
                name = f'{key}-{operator}-dRZ-{geometry}'
                self.add_data(
                    key=name,
                    data=opmat[2],
                    ref=ref,
                    units='',
                    name=operator,
                    dim=dim,
                )
            else:
                msg = "Unknown opmat type!"
                raise Exception(msg)

        # return
        if returnas is True:
            return opmat, operator, geometry, dim, ref, crop

    # -----------------
    # interp tools
    # ------------------

    def get_sample_mesh(
        self,
        key=None,
        res=None,
        grid=None,
        mode=None,
        R=None,
        Z=None,
        DR=None,
        DZ=None,
        imshow=None,
    ):
        """ Return a sampled version of the chosen mesh """
        return _mesh_comp.sample_mesh(
            mesh=self,
            key=key,
            res=res,
            grid=grid,
            mode=mode,
            R=R,
            Z=Z,
            DR=DR,
            DZ=DZ,
            imshow=imshow,
        )

    """
    def get_sample_bspline(self, key=None, res=None, grid=None, mode=None):
        return _mesh_comp.sample_bsplines(
            mesh=self,
            key=key,
            res=res,
            grid=grid,
            mode=mode,
        )
    """

    def interp2d(
        self,
        key=None,
        R=None,
        Z=None,
        grid=None,
        indbs=None,
        indt=None,
        details=None,
        reshape=None,
        res=None,
        coefs=None,
        crop=None,
        nan0=None,
        imshow=None,
    ):
        """ Interp desired data on pts """
        return _mesh_comp.interp2d(
            mesh=self,
            key=key,
            R=R,
            Z=Z,
            grid=grid,
            indbs=indbs,
            indt=indt,
            details=details,
            reshape=reshape,
            res=res,
            coefs=coefs,
            crop=crop,
            nan0=nan0,
            imshow=imshow,
        )

    # -----------------
    # crop
    # ------------------

    def crop(self, key=None, crop=None, thresh_in=None):
        """ Crop a mesh using

            - a mask of bool for each mesh elements
            - a 2d (R, Z) closed polygon

        If applied on a bspline, cropping is double-checked to make sure
        all remaining bsplines have full support domain
        """
        crop, key, thresh_in = _mesh_comp.crop(
            mesh=self,
            key=key,
            crop=crop,
            thresh_in=thresh_in,
        )

        # add crop data
        keycrop = f'{key}-crop'
        self.add_data(
            key=keycrop,
            data=crop,
            ref=self.dobj['mesh'][key]['ref'],
            dim='bool',
            quant='bool',
        )

        # update obj
        self._dobj['mesh'][key]['crop'] = keycrop
        self._dobj['mesh'][key]['crop-thresh'] = thresh_in

        # also crop bsplines
        for k0 in self.dobj.get('bsplines', {}).keys():
            if self.dobj['bsplines'][k0]['mesh'] == key:
                _mesh_comp.add_cropbs_from_crop(mesh=self, keybs=k0, keym=key)

    # -----------------
    # geometry matrix
    # ------------------

    def compute_geometry_matrix(
        self,
        key=None,
        cam=None,
        res=None,
        resMode=None,
        method=None,
        crop=None,
        name=None,
        verb=None,
    ):

        dref, ddata, dobj = _matrix_comp.compute(
            mesh=self,
            key=key,
            cam=cam,
            res=res,
            resMode=resMode,
            method=method,
            crop=crop,
            name=name,
            verb=verb,
        )

        return Matrix(dref=dref, ddata=ddata, dobj=dobj)

    # -----------------
    # plotting
    # ------------------

    def plot_mesh(
        self,
        key=None,
        ind_knot=None,
        ind_cent=None,
        crop=None,
        color=None,
        dax=None,
        dmargin=None,
        fs=None,
        dleg=None,
    ):

        return _mesh_plot.plot_mesh(
            mesh=self,
            key=key,
            ind_knot=ind_knot,
            ind_cent=ind_cent,
            crop=crop,
            color=color,
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dleg=dleg,
        )

    def plot_bsplines(
        self,
        key=None,
        ind=None,
        knots=None,
        cents=None,
        res=None,
        plot_mesh=None,
        cmap=None,
        dax=None,
        dmargin=None,
        fs=None,
        dleg=None,
    ):

        return _mesh_plot.plot_bspline(
            mesh=self,
            key=key,
            ind=ind,
            knots=knots,
            cents=cents,
            res=res,
            plot_mesh=plot_mesh,
            cmap=cmap,
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dleg=dleg,
        )

    def plot_profile2d(
        self,
        key=None,
        coefs=None,
        indt=None,
        res=None,
        vmin=None,
        vmax=None,
        cmap=None,
        dax=None,
        dmargin=None,
        fs=None,
        dcolorbar=None,
        dleg=None,
    ):
        return _mesh_plot.plot_profile2d(
            mesh=self,
            key=key,
            coefs=coefs,
            indt=indt,
            res=res,
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dcolorbar=dcolorbar,
            dleg=dleg,
        )


class Matrix(Mesh2DRect):

    def plot_geometry_matrix(
        self,
        cam=None,
        key=None,
        indbf=None,
        indchan=None,
        vmin=None,
        vmax=None,
        res=None,
        cmap=None,
        dax=None,
        dmargin=None,
        fs=None,
        dcolorbar=None,
        dleg=None,
    ):
        return _matrix_plot.plot_geometry_matrix(
            cam=cam,
            matrix=self,
            key=key,
            indbf=indbf,
            indchan=indchan,
            vmin=vmin,
            vmax=vmax,
            res=res,
            cmap=cmap,
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dcolorbar=dcolorbar,
            dleg=dleg,
        )



# -*- coding: utf-8 -*-


# Built-in
import warnings

# Common
import numpy as np
from matplotlib.path import Path


# tofu
from . import _generic_check
from . import _mesh_checks
from . import _mesh_bsplines


# #############################################################################
# #############################################################################
#                           Mesh2DRect - select
# #############################################################################


def _select_ind(
    mesh=None,
    key=None,
    ind=None,
    elements=None,
    returnas=None,
    crop=None,
):
    """ ind can be:
            - None
            - tuple: (R, Z), possibly 2d
            - 'tuple-flat': (R, Z) flattened
            - np.ndarray: array of unique indices
            - 'array-flat': flattened ordered array of unique
    """

    # ------------
    # check inputs

    ind, elements, returnas, crop = _mesh_checks._select_ind_check(
        ind=ind,
        elements=elements,
        returnas=returnas,
        crop=crop,
    )

    lk1 = list(mesh.dobj[mesh._groupmesh].keys())
    lk2 = list(mesh.dobj.get('bsplines', {}).keys())
    if key is None and len(lk1 + lk2) == 1:
        key = (lk1 + lk2)[0]
    if key not in lk1 + lk2:
        msg = (
            "Arg key must be a valid mesh identifier!\n"
            f"\t available: {lk1+lk2}\n"
            f"\t- provided: {key}"
        )
        raise Exception(msg)

    cat = 'mesh' if key in lk1 else 'bsplines'
    elem = f'{elements}' if key in lk1 else 'ref'
    kR, kZ = mesh.dobj[cat][key][elem]

    nR = mesh.ddata[kR]['data'].size
    nZ = mesh.ddata[kZ]['data'].size

    # ------------
    # ind to tuple

    ind_bool = np.zeros((nR, nZ), dtype=bool)
    if ind is None:
        # make sure R is varying in dimension 0
        ind_tup = (
            np.repeat(np.arange(0, nR)[:, None], nZ, axis=1),
            np.tile(np.arange(0, nZ), (nR, 1)),
        )
        ind_bool[...] = True

    elif isinstance(ind, tuple):
        c0 = (
            np.all((ind[0] >= 0) & (ind[0] < nR))
            and np.all((ind[1] >= 0) & (ind[1] < nZ))
        )
        if not c0:
            msg = f"Arg ind has non-valid values (< 0 or >= size ({nR}, {nZ}))"
            raise Exception(msg)
        ind_tup = ind
        ind_bool[ind_tup[0], ind_tup[1]] = True

    else:
        if np.issubdtype(ind.dtype, np.integer):
            c0 = np.all((ind >= 0) & (ind < nR*nZ))
            if not c0:
                msg = (
                    f"Arg ind has non-valid values (< 0 or >= size ({nR*nZ}))"
                )
                raise Exception(msg)
            ind_tup = (ind % nR, ind // nR)
            ind_bool[ind_tup[0], ind_tup[1]] = True
        elif np.issubdtype(ind.dtype, np.bool_):
            if ind.shape != (nR, nZ):
                msg = (
                    f"Arg ind, when array of bool, must have shape {(nR,nZ)}\n"
                    f"Provided: {ind.shape}"
                )
                raise Exception(msg)
            # make sure R varies first
            ind_tup = ind.T.nonzero()[::-1]
            ind_bool = ind
        else:
            msg = f"Unknown ind dtype!\n\t- ind.dtype: {ind.dtype}"
            raise Exception(msg)

    if ind_tup[0].shape != ind_tup[1].shape:
        msg = (
            "ind_tup components do not have the same shape!\n"
            f"\t- ind_tup[0].shape = {ind_tup[0].shape}\n"
            f"\t- ind_tup[1].shape = {ind_tup[1].shape}"
        )
        raise Exception(msg)

    # ------------
    # optional crop

    crop = (
        crop is True
        and mesh.dobj[cat][key]['crop'] is not False
        and bool(np.any(~mesh.ddata[mesh.dobj[cat][key]['crop']]['data']))
    )
    if crop is True:
        cropi = mesh.ddata[mesh.dobj[cat][key]['crop']]['data']
        if cat == 'mesh' and elements == 'knots':
            raise NotImplementedError()
        elif ind_tup[0].shape == cropi.shape:
            ind_bool = ind_bool & cropi
            # ind_tup is not 2d anymore
            ind_tup = ind_bool.T.nonzero()[::-1]  # R varies first
            warnings.warn("ind is not 2d anymore!")
        else:
            ind_bool = ind_bool & cropi
            ind_tup = ind_bool.T.nonzero()[::-1]

    # ------------
    # tuple to return

    if returnas is tuple:
        out = ind_tup
    elif returnas == 'tuple-flat':
        # make sure R is varying first
        out = (ind_tup[0].T.ravel(), ind_tup[1].T.ravel())
    elif returnas is np.ndarray:
        out = ind_tup[0] + ind_tup[1]*nR
    elif returnas == 'array-flat':
        # make sure R is varying first
        out = (ind_tup[0] + ind_tup[1]*nR).T.ravel()
    else:
        out = ind_bool

    return out


def _select_mesh(
    mesh=None,
    key=None,
    ind=None,
    elements=None,
    returnas=None,
    return_neighbours=None,
):
    """ ind is a tuple """

    # ------------
    # check inputs

    elements, returnas, return_neighbours = _mesh_checks._select_check(
        elements=elements,
        returnas=returnas,
        return_neighbours=return_neighbours,
    )

    lk = list(mesh.dobj[mesh._groupmesh])
    if key is None and len(lk) == 1:
        key = lk[0]
    if key not in lk:
        msg = (
            "Arg key must be a valid mesh identifier!\n"
            f"\t available: {lk}\n"
            f"\t- provided: {key}"
        )
        raise Exception(msg)

    # ------------
    # prepare

    kR, kZ = mesh.dobj[mesh._groupmesh][key][f'{elements}']
    R = mesh.ddata[kR]['data']
    Z = mesh.ddata[kZ]['data']
    nR = R.size
    nZ = Z.size

    # ------------
    # non-trivial case

    if returnas == 'ind':
        out = ind
    else:
        out = R[ind[0]], Z[ind[1]]

    # ------------
    # neighbours

    if return_neighbours is True:

        elneig = 'cents' if elements == 'knots' else 'knots'
        kRneig, kZneig = mesh.dobj[mesh._groupmesh][key][f'{elneig}']
        Rneig = mesh.ddata[kRneig]['data']
        Zneig = mesh.ddata[kZneig]['data']
        nRneig = Rneig.size
        nZneig = Zneig.size

        # get tuple indices of neighbours
        shape = tuple(np.r_[ind[0].shape, 4])
        neig = (
            np.zeros(shape, dtype=int),
            np.zeros(shape, dtype=int),
        )
        rsh = tuple(
            [4 if ii == len(shape)-1 else 1 for ii in range(len(shape))]
        )

        if elements == 'cents':
            neig[0][...] = ind[0][..., None] + np.r_[0, 1, 1, 0].reshape(rsh)
            neig[1][...] = ind[1][..., None] + np.r_[0, 0, 1, 1].reshape(rsh)
        elif elements == 'knots':
            neig[0][...] = ind[0][..., None] + np.r_[-1, 0, 0, -1].reshape(rsh)
            neig[1][...] = ind[1][..., None] + np.r_[-1, -1, 0, 0].reshape(rsh)
            neig[0][(neig[0] < 0) | (neig[0] >= nRneig)] = -1
            neig[1][(neig[1] < 0) | (neig[1] >= nZneig)] = -1

        # return neighbours in desired format
        if returnas == 'ind':
            neig_out = neig
        else:
            neig_out = np.array([
                Rneig[neig[0]], Zneig[neig[1]],
            ])
            neig_out[:, (neig[0] == -1) | (neig[1] == -1)] = np.nan

        return out, neig_out
    else:
        return out


def _select_bsplines(
    mesh=None,
    key=None,
    ind=None,
    returnas=None,
    return_cents=None,
    return_knots=None,
):
    """ ind is a tuple """

    # ------------
    # check inputs

    _, returnas, _ = _mesh_checks._select_check(
        returnas=returnas,
    )

    lk = list(mesh.dobj['bsplines'])
    if key is None and len(lk) == 1:
        key = lk[0]
    if key not in lk:
        msg = (
            "Arg key must be a valid mesh identifier!\n"
            f"\t available: {lk}\n"
            f"\t- provided: {key}"
        )
        raise Exception(msg)

    keym = mesh.dobj['bsplines'][key]['mesh']
    kRk, kZk = mesh.dobj['mesh'][keym]['knots']
    kRc, kZc = mesh.dobj['mesh'][keym]['cents']

    # ------------
    # knots, cents

    out = _mesh2DRect_bsplines_knotscents(
        returnas=returnas,
        return_knots=return_knots,
        return_cents=return_cents,
        ind=ind,
        deg=mesh.dobj['bsplines'][key]['deg'],
        Rknots=mesh.ddata[kRk]['data'],
        Zknots=mesh.ddata[kZk]['data'],
        Rcents=mesh.ddata[kRc]['data'],
        Zcents=mesh.ddata[kZc]['data'],
    )

    # ------------
    # return

    if return_cents is True and return_knots is True:
        return ind, out[0], out[1]
    elif return_cents is True or return_knots is True:
        return ind, out
    else:
        return ind


# #############################################################################
# #############################################################################
#                           Mesh2DRect - bsplines
# #############################################################################


def _mesh2DRect_bsplines(mesh=None, keym=None, keybs=None, deg=None):

    # --------------
    # create bsplines

    kR, kZ = mesh.dobj['mesh'][keym]['knots']
    Rknots = mesh.ddata[kR]['data']
    Zknots = mesh.ddata[kZ]['data']

    kRbsc = f'{keybs}-R'
    kZbsc = f'{keybs}-Z'

    (
        shapebs, Rbs_cent, Zbs_cent,
        knots_per_bs_R, knots_per_bs_Z,
    ) = _mesh_bsplines.get_bs2d_RZ(
        deg=deg, Rknots=Rknots, Zknots=Zknots,
    )
    nbs = int(np.prod(shapebs))

    func_details, func_sum, clas = _mesh_bsplines.get_bs2d_func(
        deg=deg,
        Rknots=Rknots,
        Zknots=Zknots,
        shapebs=shapebs,
        knots_per_bs_R=knots_per_bs_R,
        knots_per_bs_Z=knots_per_bs_Z,
    )
    ref = (kRbsc, kZbsc)

    # ----------------
    # format into dict

    dref = {
        kRbsc: {
            'data': Rbs_cent,
            'units': 'm',
            'dim': 'distance',
            'quant': 'R',
            'name': 'R',
            'group': 'R',
        },
        kZbsc: {
            'data': Zbs_cent,
            'units': 'm',
            'dim': 'distance',
            'quant': 'Z',
            'name': 'Z',
            'group': 'Z',
        },
        keybs: {
            'data': np.arange(0, nbs),
            'units': '',
            'dim': 'index',
            'quant': 'index',
            'name': '',
            'group': 'index',
        },
    }

    dobj = {
        'bsplines': {
            keybs: {
                'deg': deg,
                'mesh': keym,
                'ref': ref,
                'shape': shapebs,
                'crop': False,
                'func_details': func_details,
                'func_sum': func_sum,
                'class': clas,
            }
        },
    }

    return dref, dobj


def add_cropbs_from_crop(mesh=None, keybs=None, keym=None):

    # ----------------
    # get

    kcropbs = False
    if mesh.dobj['mesh'][keym]['crop'] is not False:
        kcropm = mesh.dobj['mesh'][keym]['crop']
        cropbs = _get_cropbs_from_crop(
            mesh=mesh,
            crop=mesh.ddata[kcropm]['data'],
            keybs=keybs,
        )
        kcropbs = f'{keybs}-crop'
        kcroppedbs = f'{keybs}-cropped'

    # ----------------
    # optional crop

    if kcropbs is not False:

        # add cropped flat reference
        mesh.add_ref(
            key=kcroppedbs,
            data=np.arange(0, cropbs.sum()),
            group='index',
        )

        mesh.add_data(
            key=kcropbs,
            data=cropbs,
            ref=mesh._dobj['bsplines'][keybs]['ref'],
            dim='bool',
            quant='bool',
        )
        mesh._dobj['bsplines'][keybs]['crop'] = kcropbs


def _mesh2DRect_bsplines_knotscents(
    returnas=None,
    return_knots=None,
    return_cents=None,
    ind=None,
    deg=None,
    Rknots=None,
    Zknots=None,
    Rcents=None,
    Zcents=None,
):

    # -------------
    # check inputs

    return_knots = _generic_check._check_var(
        return_knots, 'return_knots',
        types=bool,
        default=True,
    )
    return_cents = _generic_check._check_var(
        return_cents, 'return_cents',
        types=bool,
        default=True,
    )
    if return_knots is False and return_cents is False:
        return

    # -------------
    # compute

    if return_knots is True:

        knots_per_bs_R = _mesh_bsplines._get_bs2d_func_knots(
            Rknots, deg=deg, returnas=returnas,
        )
        knots_per_bs_Z = _mesh_bsplines._get_bs2d_func_knots(
            Zknots, deg=deg, returnas=returnas,
        )
        if ind is not None:
            knots_per_bs_R = knots_per_bs_R[:, ind[0]]
            knots_per_bs_Z = knots_per_bs_Z[:, ind[1]]

        nknots = knots_per_bs_R.shape[0]
        knots_per_bs_R = np.tile(knots_per_bs_R, (nknots, 1))
        knots_per_bs_Z = np.repeat(knots_per_bs_Z, nknots, axis=0)

    if return_cents is True:

        cents_per_bs_R = _mesh_bsplines._get_bs2d_func_cents(
            Rcents, deg=deg, returnas=returnas,
        )
        cents_per_bs_Z = _mesh_bsplines._get_bs2d_func_cents(
            Zcents, deg=deg, returnas=returnas,
        )
        if ind is not None:
            cents_per_bs_R = cents_per_bs_R[:, ind[0]]
            cents_per_bs_Z = cents_per_bs_Z[:, ind[1]]

        ncents = cents_per_bs_R.shape[0]
        cents_per_bs_R = np.tile(cents_per_bs_R, (ncents, 1))
        cents_per_bs_Z = np.repeat(cents_per_bs_Z, ncents, axis=0)

    # -------------
    # return

    if return_knots is True and return_cents is True:
        out = (
            (knots_per_bs_R, knots_per_bs_Z), (cents_per_bs_R, cents_per_bs_Z)
        )
    elif return_knots is True:
        out = (knots_per_bs_R, knots_per_bs_Z)
    else:
        out = (cents_per_bs_R, cents_per_bs_Z)
    return out


# #############################################################################
# #############################################################################
#                           Mesh2DRect - sample
# #############################################################################


def _sample_mesh_check(
    mesh=None,
    key=None,
    res=None,
    mode=None,
    grid=None,
    imshow=None,
    R=None,
    Z=None,
    DR=None,
    DZ=None,
):

    # -----------
    # Parameters

    # key
    lk = list(mesh.dobj[mesh._groupmesh].keys())
    if key is None and len(lk) == 1:
        key = lk[0]
    if key not in lk:
        msg = (
            "Arg key must point to a valid mesh id!\n"
            f"\t- available: {lk}\n"
            f"\t- provided: {key}\n"
        )
        raise Exception(msg)

    # res
    if res is None:
        res = 0.1
    if np.isscalar(res):
        res = [res, res]
    c0 = (
        isinstance(res, list)
        and len(res) == 2
        and all([np.isscalar(rr) and rr > 0 for rr in res])
    )
    if not c0:
        msg = f"Arg res must be a list of 2 positive floats!\nProvided: {res}"
        raise Exception(msg)

    # mode
    mode = _generic_check._check_var(
        mode, 'mode',
        types=str,
        default='abs',
    )

    # grid
    grid = _generic_check._check_var(
        grid, 'grid',
        types=bool,
        default=False,
    )

    # imshow
    imshow = _generic_check._check_var(
        imshow, 'imshow',
        types=bool,
        default=False,
    )

    # R, Z
    if R is None and Z is None:
        pass
    elif R is None and np.isscalar(Z):
        pass
    elif Z is None and np.isscalar(R):
        pass
    else:
        msg = (
            "For mesh discretisation, (R, Z) can be either:\n"
            "\t- (None, None): will be created\n"
            "\t- (scalar, None): A vertical line will be created\n"
            "\t- (None, scalar): A horizontal line will be created\n"
        )
        raise Exception(msg)

    # -------------
    # R, Z

    kR, kZ = mesh.dobj[mesh._groupmesh][key]['knots']
    Rk = mesh.ddata[kR]['data']
    Zk = mesh.ddata[kZ]['data']

    # custom R xor Z for vertical / horizontal lines only
    if R is None and Z is not None:
        R = Rk
    if Z is None and R is not None:
        Z = Zk

    # custom DR or DZ for mode='abs' only
    if DR is not None or DZ is not None:
        if mode != 'abs':
            msg = "Custom DR or DZ can only be provided with mode = 'abs'!"
            raise Exception(msg)

        for DD, DN in [(DR, 'DR'), (DZ, 'DZ')]:
            if DD is not None:
                c0 = (
                    hasattr(DD, '__iter__')
                    and len(DD) == 2
                    and all([
                        rr is None or (np.isscalar(rr) and np.isfinite(rr))
                        for rr in DD
                    ])
                )
                if not c0:
                    msg = f'Arg {DN} must be an iterable of 2 scalars!'
                    raise Exception(msg)

    if DR is None:
        DR = [Rk.min(), Rk.max()]
    if DZ is None:
        DZ = [Zk.min(), Zk.max()]

    return key, res, mode, grid, imshow, R, Z, DR, DZ, Rk, Zk


def sample_mesh(
    mesh=None,
    key=None,
    res=None,
    mode=None,
    R=None,
    Z=None,
    DR=None,
    DZ=None,
    grid=None,
    imshow=None,
):

    # -------------
    # check inputs

    key, res, mode, grid, imshow, R, Z, DR, DZ, Rk, Zk = _sample_mesh_check(
        mesh=mesh,
        key=key,
        res=res,
        mode=mode,
        grid=grid,
        imshow=imshow,
        R=R,
        Z=Z,
        DR=DR,
        DZ=DZ,
    )

    # -------------
    # compute

    if mode == 'abs':
        if R is None:
            nR = int(np.ceil((DR[1] - DR[0]) / res[0]))
            R = np.linspace(DR[0], DR[1], nR)
        if Z is None:
            nZ = int(np.ceil((DZ[1] - DZ[0]) / res[1]))
            Z = np.linspace(DZ[0], DZ[1], nZ)
    else:
        if R is None:
            nR = int(np.ceil(1./res[0]))
            kR = np.linspace(0, 1, nR, endpoint=False)[None, :]
            R = np.concatenate((
                (Rk[:-1, None] + kR*np.diff(Rk)[:, None]).ravel(),
                Rk[-1:],
            ))
        if Z is None:
            nZ = int(np.ceil(1./res[1]))
            kZ = np.linspace(0, 1, nZ, endpoint=False)[None, :]
            Z = np.concatenate((
                (Zk[:-1, None] + kZ*np.diff(Zk)[:, None]).ravel(),
                Zk[-1:],
            ))

    if np.isscalar(R):
        R = np.full(Z.shape, R)
    if np.isscalar(Z):
        Z = np.full(R.shape, Z)

    # ------------
    # grid

    if grid is True:
        nZ = Z.size
        nR = R.size
        if imshow is True:
            R = np.tile(R, (nZ, 1))
            Z = np.repeat(Z[:, None], nR, axis=1)
        else:
            R = np.repeat(R[:, None], nZ, axis=1)
            Z = np.tile(Z, (nR, 1))

    return R, Z


# #############################################################################
# #############################################################################
#                           Mesh2DRect - crop
# #############################################################################


def _crop_check(mesh=None, key=None, crop=None, thresh_in=None):

    # key
    lkm = list(mesh.dobj['mesh'].keys())
    key = _generic_check._check_var(
        key, 'key',
        default=None,
        types=str,
        allowed=lkm,
    )

    shape = mesh.dobj['mesh'][key]['shape']

    # crop
    c0 = (
        isinstance(crop, np.ndarray)
        and crop.ndim == 2
        and np.all(np.isfinite(crop))
        and (
            (
                crop.shape[0] == 2
                and np.allclose(crop[:, 0], crop[:, -1])
                and (
                    np.issubdtype(crop.dtype, np.integer)
                    or np.issubdtype(crop.dtype, np.floating)
                )
            )
            or (
                crop.shape == shape
                and crop.dtype == np.bool_
            )
        )
    )
    if not c0:
        msg = (
            "Arg crop must be either:\n"
            f"\t- array of bool: mask of shape {shape}\n"
            f"\t- array of floats: (2, npts) closed (R, Z) polygon\n"
            f"Provided:\n{crop}"
        )
        raise Exception(msg)

    cropbool = crop.dtype == np.bool_

    # thresh_in
    if thresh_in is None:
        thresh_in = 3
    maxth = 5 if mesh.dobj['mesh'][key]['type'] == 'rect' else 4
    c0 = isinstance(thresh_in, (int, np.integer)) and (1 <= thresh_in <= maxth)
    if not c0:
        msg = (
            f"Arg thresh_in must be a int in {1} <= thresh_in <= {maxth}\n"
            f"Provided: {thresh_in}"
        )
        raise Exception(msg)

    return key, cropbool, thresh_in


def crop(mesh=None, key=None, crop=None, thresh_in=None):

    # ------------
    # check inputs

    key, cropbool, thresh_in = _crop_check(
        mesh=mesh, key=key, crop=crop, thresh_in=thresh_in,
    )

    # -----------
    # if crop is a poly => compute as bool

    if not cropbool:
        (Rc, Zc), (Rk, Zk) = mesh.select_mesh_elements(
            key=key, elements='cents', return_neighbours=True, returnas='data',
        )
        nR, nZ = Rc.shape
        npts = Rk.shape[-1] + 1

        pts = np.concatenate(
            (
                np.concatenate((Rc[:, :, None], Rk), axis=-1)[..., None],
                np.concatenate((Zc[:, :, None], Zk), axis=-1)[..., None],
            ),
            axis=-1,
        ).reshape((npts*nR*nZ, 2))

        isin = Path(crop.T).contains_points(pts).reshape((nR, nZ, npts))
        crop = np.sum(isin, axis=-1) >= thresh_in

    return crop, key, thresh_in


def _get_cropbs_from_crop(mesh=None, crop=None, keybs=None):

    if isinstance(crop, str) and crop in mesh.ddata.keys():
        crop = mesh.ddata[crop]['data']

    shaperef = mesh.dobj['mesh'][mesh.dobj['bsplines'][keybs]['mesh']]['shape']
    if crop.shape != shaperef:
        msg = "Arg crop seems to have the wrong shape!"
        raise Exception(msg)

    keym = mesh.dobj['bsplines'][keybs]['mesh']
    kRk, kZk = mesh.dobj['mesh'][keym]['knots']
    kRc, kZc = mesh.dobj['mesh'][keym]['cents']

    cents_per_bs_R, cents_per_bs_Z = _mesh2DRect_bsplines_knotscents(
        returnas='ind',
        return_knots=False,
        return_cents=True,
        ind=None,
        deg=mesh.dobj['bsplines'][keybs]['deg'],
        Rknots=mesh.ddata[kRk]['data'],
        Zknots=mesh.ddata[kZk]['data'],
        Rcents=mesh.ddata[kRc]['data'],
        Zcents=mesh.ddata[kZc]['data'],
    )

    shapebs = mesh.dobj['bsplines'][keybs]['shape']
    cropbs = np.array([
        [
            np.all(crop[cents_per_bs_R[:, ii], cents_per_bs_Z[:, jj]])
            for jj in range(shapebs[1])
        ]
        for ii in range(shapebs[0])
    ], dtype=bool)

    return cropbs


# #############################################################################
# #############################################################################
#                           Mesh2DRect - interp
# #############################################################################


def _interp_check(
    mesh=None,
    key=None,
    R=None,
    Z=None,
    grid=None,
    indbs=None,
    indt=None,
    details=None,
    res=None,
    coefs=None,
    crop=None,
    nan0=None,
    imshow=None,
):
    # key
    dk = {
        kk: [
            k1 for k1, v1 in mesh.dobj['bsplines'].items()
            if mesh.ddata[kk]['ref'][-2:] == v1['ref']
        ][0]
        for kk in mesh.ddata.keys()
        if any([
            mesh.ddata[kk]['ref'][-2:] == v1['ref']
            for v1 in mesh.dobj['bsplines'].values()
        ])
        and 'crop' not in kk
    }
    dk.update({kk: kk for kk in mesh.dobj['bsplines'].keys()})
    if key is None and len(dk) == 1:
        key = list(dk.keys())[0]
    if key not in dk.keys():
        msg = (
            "Arg key must the key to a data referenced on a bsplines set\n"
            f"\t- available: {dk.keys()}\n"
            f"\t- provided: {key}\n"
        )
        raise Exception(msg)
    keybs = dk[key]
    keym = mesh.dobj['bsplines'][keybs]['mesh']

    # coefs
    shapebs = mesh.dobj['bsplines'][keybs]['shape']
    if coefs is None:
        if key == keybs:
            pass
        else:
            coefs = mesh.ddata[key]['data']
    else:
        c0 = (
            coefs.ndim in [len(shapebs), len(shapebs) + 1]
            and coefs.shape[-2:] == shapebs
        )
        if not c0:
            msg = (
                "Arg coefs must be a {shapebs} array!\n"
                f"Provided: coefs.shape"
            )
            raise Exception(msg)

    # indbs

    # indt
    c0 = (
        indt is not None
        and coefs is not None
        and coefs.ndim == len(shapebs) + 1
    )
    if c0:
        if coefs.shape[0] == 1:
            indt = 0
        try:
            assert np.isscalar(indt) and np.isfinite(indt)
            assert indt < coefs.shape[0]
            indt = int(indt)
        except Exception as err:
            msg = (
                f"Arg indt should be a int!\nProvided: {indt}"
            )
            raise Exception(msg)
        coefs = coefs[indt:indt+1, ...]

    # details
    details = _generic_check._check_var(
        details, 'details',
        types=bool,
        default=False,
    )

    # crop
    crop = _generic_check._check_var(
        crop, 'crop',
        types=bool,
        default=True,
    )

    # nan0
    nan0 = _generic_check._check_var(
        nan0, 'nan0',
        types=bool,
        default=True,
    )

    # R, Z
    if R is None or Z is None:
        R, Z = mesh.get_sample_mesh(
            key=keym,
            res=res,
            mode='abs',
            grid=True,
            R=R,
            Z=Z,
            imshow=imshow,
        )
    else:
        if not isinstance(R, np.ndarray):
            try:
                R = np.atleast_1d(R).astype(float)
            except Exception as err:
                msg = "R must be convertible to np.arrays of floats"
                raise Exception(msg)
        if not isinstance(Z, np.ndarray):
            try:
                Z = np.atleast_1d(Z).astype(float)
            except Exception as err:
                msg = "Z must be convertible to np.arrays of floats"
                raise Exception(msg)

        # grid
        if grid is None:
            grid = R.shape != Z.shape
        if not isinstance(grid, bool):
            msg = f"Arg grid must be a bool!\nProvided: {grid}"
            raise Exception(msg)

        if grid is True and (R.ndim > 1 or Z.ndim > 1):
            msg = "If grid=True, R and Z must be 1d!"
            raise Exception(msg)
        elif grid is False and R.shape != Z.shape:
            msg = "If grid=False, R and Z must have the same shape!"
            raise Exception(msg)

        if grid is True:
            R = np.tile(R, Z.size)
            Z = np.repeat(Z, R.size)

    return key, keybs, R, Z, coefs, indbs, indt, details, crop, nan0


def interp2d(
    mesh=None,
    key=None,
    R=None,
    Z=None,
    coefs=None,
    indbs=None,
    indt=None,
    grid=None,
    details=None,
    reshape=None,
    res=None,
    crop=None,
    nan0=None,
    imshow=None,
):

    # ---------------
    # check inputs

    key, keybs, R, Z, coefs, indbs, indt, details, crop, nan0 = _interp_check(
        mesh=mesh,
        key=key,
        R=R,
        Z=Z,
        coefs=coefs,
        indbs=indbs,
        indt=indt,
        grid=grid,
        details=details,
        res=res,
        crop=crop,
        nan0=nan0,
        imshow=imshow,
    )

    # ---------------
    # prepare

    if details is True:
        fname = 'func_details'
    elif details is False:
        fname = 'func_sum'
    else:
        raise Exception("Unknown details!")

    # ---------------
    # interp

    cropbs = mesh.dobj['bsplines'][keybs]['crop']
    if cropbs is not False:
        cropbs = mesh.ddata[cropbs]['data']

    if details is not False:
        indbs_tuple_flat = mesh.select_ind(
            key=keybs,
            returnas='tuple-flat',
            ind=indbs,
        )
    else:
        indbs_tuple_flat = None

    val = mesh.dobj['bsplines'][keybs][fname](
        R, Z,
        coefs=coefs,
        crop=crop,
        cropbs=cropbs,
        indbs_tuple_flat=indbs_tuple_flat,
        reshape=reshape,
    )

    # ---------------
    # post-treatment

    if nan0 is True:
        val[val == 0] = np.nan

    return val


# #############################################################################
# #############################################################################
#                           Mesh2DRect - operators
# #############################################################################


def get_bsplines_operator(
    coll,
    key=None,
    operator=None,
    geometry=None,
    crop=None,
    store=None,
    returnas=None,
):

    # check inputs
    lk = list(coll.dobj.get('bsplines', {}).keys())
    key = _generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lk,
    )

    store = _generic_check._check_var(
        store, 'store',
        default=True,
        types=bool,
    )

    returnas = _generic_check._check_var(
        returnas, 'returnas',
        default=store is False,
        types=bool,
    )

    crop = _generic_check._check_var(
        crop, 'crop',
        default=True,
        types=bool,
    )

    cropbs_flat = coll.dobj['bsplines'][key]['crop']
    if cropbs_flat is not False and crop is True:
        cropbs_flat = coll.ddata[cropbs_flat]['data'].ravel(order='F')
        keycropped = f'{key}-cropped'
    else:
        crop = False
        cropbs_flat = False
        keycropped = key

    # compute and return
    (
        opmat, operator, geometry, dim,
    ) = coll.dobj['bsplines'][key]['class'].get_operator(
        operator=operator,
        geometry=geometry,
        cropbs_flat=cropbs_flat,
    )

    # cropping
    if operator == 'D0':
        ref = (keycropped,)
    elif 'N' in operator:
        ref = (keycropped, keycropped)

    return opmat, operator, geometry, dim, ref, crop, store, returnas, key



# -*- coding: utf-8 -*-


# Built-in
import copy


# Common
import numpy as np


# tofu
from . import _generic_check
from . import _mesh_bsplines


# #############################################################################
# #############################################################################
#                           Matrix - compute
# #############################################################################


def _compute_check(
    mesh=None,
    key=None,
    method=None,
    resMode=None,
    crop=None,
    name=None,
    verb=None,
):

    # key
    lk = list(mesh.dobj.get('bsplines', {}).keys())
    if key is None and len(lk) == 1:
        key = lk[0]
    if key not in lk:
        msg = (
            "Arg key must be a valid bspline identifier!\n"
            f"\t- available: {lk}\n"
            f"\t- provided:  {key}"
        )
        raise Exception(msg)

    # method
    method = _generic_check._check_var(
        method, 'method',
        default='los',
        types=str,
        allowed=['los'],
    )

    # resMode
    resMode = _generic_check._check_var(
        resMode, 'resMode',
        default='abs',
        types=str,
        allowed=['abs', 'rel'],
    )

    # crop
    crop = _generic_check._check_var(
        crop, 'crop',
        default=True,
        types=bool,
    )
    crop = crop and mesh.dobj['bsplines'][key]['crop'] is not False

    # name
    if name is None:
        lmat = [
            kk for kk in mesh.dobj.get('matrix', {}).keys()
            if kk.startswith('matrix')
        ]
        name = f'matrix{len(lmat)}'
    c0 = (
        isinstance(name, str)
        and name not in mesh.dobj.get('matrix', {}).keys()
    )
    if not c0:
        msg = (
            "Arg name must be a str not already taken!\n"
            f"\t- already taken: {mesh.dobj.get('matrix', {}).keys()}\n"
            f"\t- provided: {name}"
        )
        raise Exception(msg)

    # verb
    if verb is None:
        verb = True
    if not isinstance(verb, bool):
        msg = (
            f"Arg verb must be a bool!\n"
            f"\t- provided: {verb}"
        )
        raise Exception(msg)

    return key, method, resMode, crop, name, verb


def compute(
    mesh=None,
    key=None,
    cam=None,
    res=None,
    resMode=None,
    method=None,
    crop=None,
    name=None,
    verb=None,
):
    """ Compute the geometry matrix using:
            - a mesh2DRect instance with a key to a bspline set
            - a cam instance with a resolution
    """

    # -----------
    # check input

    key, method, resMode, crop, name, verb = _compute_check(
        mesh=mesh, key=key, method=method, resMode=resMode,
        crop=crop, name=name, verb=verb,
    )

    # -----------
    # prepare

    nlos = cam.nRays
    shapebs = mesh.dobj['bsplines'][key]['shape']

    # -----------
    # compute

    if method == 'los':
        # discretize lines once, then evaluated at points
        pts, reseff, ind = cam.get_sample(
            res=res,
            resMode=resMode,
            DL=None,
            method='sum',
            ind=None,
            pts=True,
            compact=True,
            num_threads=10,
            Test=True,
        )
        lr = np.split(np.hypot(pts[0, :], pts[1, :]), ind)
        lz = np.split(pts[2, :], ind)

        if verb:
            nmax = len(f"Geometry matrix, channel {nlos} / {nlos}")
            nn = 10**(np.log10(nlos)-1)

        # prepare indices
        indbs = mesh.select_ind(
            key=key,
            returnas=tuple,
            crop=crop,
        )

        # prepare matrix
        shapemat = tuple(np.r_[nlos, indbs[0].size])
        mat = np.zeros(shapemat, dtype=float)

        for ii in range(nlos):

            # verb
            if verb:
                msg = f"Geom. matrix, chan {ii+1} / {nlos}".ljust(nmax)
                end = '\n' if ii == nlos-1 else '\r'
                print(msg, end=end, flush=True)

            # compute
            mat[ii, :] = np.nansum(
                mesh.interp2d(
                    key=key,
                    R=lr[ii],
                    Z=lz[ii],
                    grid=False,
                    indbs=indbs,
                    details=True,
                    reshape=False,
                ),
                axis=0,
            )

        mat = mat * reseff[:, None]
        # scpintg.simps(val, x=None, axis=-1, dx=loc_eff_res[0])

    # -----------
    # return

    # extract existing parts relevant to the geometry matrix (mesh + bsplines)

    # dobj
    km = mesh.dobj['bsplines'][key]['mesh']
    dobj = copy.deepcopy({
        'mesh': {
            km: mesh.dobj['mesh'][km],
        },
        'bsplines': {
            key: mesh.dobj['bsplines'][key],
        },
    })

    # dref
    keycropped = f'{key}-cropped' if crop is True else key
    lref = (
        list(mesh.dobj['mesh'][km]['cents'])
        + list(mesh.dobj['mesh'][km]['knots'])
        + list(mesh.dobj['bsplines'][key]['ref'])
        + [key]
    )
    if crop is True:
        lref.append(keycropped)

    dref = {k0: mesh.dref[k0] for k0 in lref}
    for k0 in lref:
        dref[k0].update({
            k1: v1 for k1, v1 in mesh.ddata[k0].items()
            if k1 not in ['ref', 'group']
        })
    dref = copy.deepcopy(dref)

    # ddata
    lcrop = [mesh.dobj['mesh'][km]['crop'], mesh.dobj['bsplines'][key]['crop']]
    ddata = copy.deepcopy({
        k0: v0 for k0, v0 in mesh.ddata.items()
        if k0 in lcrop
    })

    # add new parts relevant to the geometry matrix (matrix + ref)
    dref.update({
        'channels': {
            'data': np.arange(0, nlos),
            'group': 'chan',
        },
    })

    ddata.update({
        name: {
            'data': mat,
            'ref': ('channels', keycropped)
        },
    })

    dobj.update({
        'matrix': {
            name: {
                'bsplines': key,
                'cam': cam.Id.Name,
                'data': name,
                'crop': crop,
                'shape': mat.shape,
            },
        },
    })

    return dref, ddata, dobj
